/*Copyright (c) 2004, Stanford University
  All rights reserved.*/

#include "clip.h"
#include <GL/glut.h>
#include <assert.h>
#include <iostream>
#include <math.h>
#include <stdlib.h>
extern "C" {    
#include "vec.h"
}   

using namespace std;
typedef float myfloat;

// WINDOW DIMENSIONS
void GetWindowDim (GLint * w, GLint * h)
{
        GLint viewport[4];
        glGetIntegerv (GL_VIEWPORT, viewport);
        *w = viewport[2];
        *h = viewport[3];
} 

//*****************************
// EYE SPACE SLICED CUBE CLASS
//*****************************
EyeSlicedCube::EyeSlicedCube (float xmin, float xmax, float ymin,
                              float ymax, float zmin, float zmax,
                              float t) 
{
        this->xmin = xmin;
        this->xmax = xmax;
        this->ymin = ymin;
        this->ymax = ymax;
        this->zmin = zmin;
        this->zmax = zmax;
        setSliceThickness (t);
        renderClipPlanes = false;
        initialized = false;
} 

EyeSlicedCube::EyeSlicedCube (float t) 
{
        this->xmin = 0;
        this->xmax = 1;
        this->ymin = 0;
        this->ymax = 1;
        this->zmin = 0;
        this->zmax = 1;
        setSliceThickness (t);
        renderClipPlanes = false;
        initialized = false;
} 

void EyeSlicedCube::setSliceThickness (float t) 
{
        thickness = t;
}

float EyeSlicedCube::getSliceThickness () 
{
        return thickness;
}
void EyeSlicedCube::addClipPlane (const Plane & clipPlane) 
{
        userClipPlanes.push_back (clipPlane);
} 

void EyeSlicedCube::removeClipPlane (int index) 
{
        vector < Plane >::iterator iter = userClipPlanes.begin ();
        iter += index;
        userClipPlanes.erase (iter);
}

const Plane & EyeSlicedCube::getClipPlane (int index) 
{
        return userClipPlanes[index];
}

static matrix4_t GetModelView () 
{
        matrix4_t modelview;
        glGetFloatv (GL_MODELVIEW_MATRIX, (float *) &modelview);
        return modelview;
}
static BoundingBox GetBoundingBox (PointList & points,  matrix4_t & t) 
{
        PointList transformPoints;
        assert (points.size () >= 1);
        
        // Transform into eye-space
        for (PointList::iterator iter = points.begin (); 
             iter != points.end (); ++iter)
        {
                Point eyePoint;
                matrix_multiply_vec_4 ((float *) &eyePoint, &t, 
                                       (float *) &(*iter));
                transformPoints.push_back (eyePoint);
        } 
        
        float minx = 0, miny = 0, minz = 0, maxx = 0, maxy = 0, maxz = 0;
        for (PointList::iterator i = transformPoints.begin (); 
             i != transformPoints.end (); ++i)
        {
                if (i == transformPoints.begin ())
                {
                        minx = maxx = i->X ();
                        miny = maxy = i->Y ();
                        minz = maxz = i->Z ();
                } 
                else
                {
                        if (i->X () < minx)
                                minx = i->X ();
                        if (i->Y () < miny)
                                miny = i->Y ();
                        if (i->Z () < minz)
                                minz = i->Z ();
                        if (i->X () > maxx)
                                maxx = i->X ();
                        if (i->Y () > maxy)
                                maxy = i->Y ();
                        if (i->Z () > maxz)
                                maxz = i->Z ();
                }
        }
        return BoundingBox (minx, miny, minz, maxx - minx, maxy - miny,
                            maxz - minz);
}

static BoundingBox GetEyeSpaceBoundingBox (matrix4_t modelview,
                                           PointList & points) 
{
        return GetBoundingBox (points, modelview);
}

static void emitPolygon (const ConvexPolygon & poly) 
{
        if (poly.verts ().size () >= 3)
        {
                for (unsigned int i = 0; i < poly.verts ().size (); i++)
                {
                        glVertex4fv ((float *) &poly.verts ()[i]);
                } 
        }
} 

static inline void emitTriangles (const ConvexPolygon & poly) 
{
        if (poly.verts ().size () >= 3)
        {
                for (unsigned int i = 2; i < poly.verts ().size (); i++)
                {
                        glColor3f (1, 1, 1);
                        glVertex4fv ((float *) &poly.verts ()[0]);
                        glVertex4fv ((float *) &poly.verts ()[i - 1]);
                        glVertex4fv ((float *) &poly.verts ()[i]);
                } 
        }
} 

static inline void CalculateCoords (float p00[3], const float v00[3],
                                    const float v01[3],
                                    const float v11[3],
                                    const float v10[3], float xfrac,
                                    float yfrac) 
{
        for (int i = 0; i < 3; i++)
        {
                p00[i] = v00[i] * (1 - xfrac) * (1 - yfrac) 
                        +v01[i] * (1 - xfrac) * yfrac 
                        +v11[i] * xfrac * yfrac  +v10[i] * xfrac * (1 - yfrac);
	} 
} 

static float inline length (const float a[3], const float b[3]) 
{
        float x = a[0] - b[0];
        float y = a[1] - b[1];
        float z = a[2] - b[2];
        return sqrt (x * x + y * y + z * z);
}

static float inline length (const float a[3]) 
{
        return sqrt (a[0] * a[0] + a[1] * a[1] + a[2] * a[2]);
}


// Renders the cube with slices that are aligned with view-space 
// axes.  These axes are determined from the current GL model-view
// matrix.
void EyeSlicedCube::renderViewSliced (matrix4_t modelview) 
{
        if (!initialized)
        {
                initialized = true;
                cubeVertices.push_back (Point (xmin, ymin, zmin));
                cubeVertices.push_back (Point (xmin, ymin, zmax));
                cubeVertices.push_back (Point (xmin, ymax, zmin));
                cubeVertices.push_back (Point (xmin, ymax, zmax));
                cubeVertices.push_back (Point (xmax, ymin, zmin));
                cubeVertices.push_back (Point (xmax, ymin, zmax));
                cubeVertices.push_back (Point (xmax, ymax, zmin));
                cubeVertices.push_back (Point (xmax, ymax, zmax));
        }
        
        // eye space bbox
        BoundingBox ebbox = GetEyeSpaceBoundingBox(modelview, cubeVertices);  
        
        // We want to render from back to front in eye space.  
        // z=-1 is far in eye, space, z = 0 near
        
        float x, y, z, w, h, d;
        x = ebbox.getX ();
        y = ebbox.getY ();
        z = ebbox.getZ ();
        w = ebbox.getWidth ();
        h = ebbox.getHeight ();
        d = ebbox.getDepth ();
        
        vector < Plane > clipPlanes;		
        clipPlanes.push_back (Plane (1, 0, 0, -xmin));     // x >= xmin
        clipPlanes.push_back (Plane (-1, 0, 0, xmax));      // x <= xmax
        clipPlanes.push_back (Plane (0, 1, 0, -ymin));
        clipPlanes.push_back (Plane (0, -1, 0, ymax));
        clipPlanes.push_back (Plane (0, 0, 1, -zmin));
        clipPlanes.push_back (Plane (0, 0, -1, zmax));
        clipPlanes.insert (clipPlanes.end (), userClipPlanes.begin (),
                           userClipPlanes.end ());
        
        for (unsigned int i = 0; i < clipPlanes.size (); i++)
                clipPlanes[i].transform (modelview);
        
        /*################################################## 
          
        MCH: This has been mucked with slightly.  We don't seem to
        need to compute eyeThickness, but I'm not positive about
        zstart.
        
        ###################################################*/
        
        // Make sure that the first slices is placed correctly so that
        // the collection of slices that fills space contains one that
        // passes through the origin.  This allows several different
        // cubes to be rendered on different machines and composited
        // such that the slicing planes match up on the cube sides
        // that they share.  
        
        //float zstart = (int) (z / thickness) * thickness;
        
        //the first slice place should go through the voxel center
        float zstart = z;
        
        // The thickness that we are given is in object space.  We want 
        // to space the slices in eye space when iterating over k below, 
        // so that the spacing of the slices in object space is of the 
        // requested thickness.
        /*float eyeThickness;
          {
          float unit_z[3] = { 0, 0, 1 };
          float unit_z_trans[3];
          matrix3_t affine_modelview;
          matrix_affine_4(&affine_modelview, &modelview);
          matrix_multiply_vec_3(unit_z_trans, &affine_modelview, unit_z);
          eyeThickness = length(unit_z_trans) * thickness;
          } */
	
        ConvexPolygon poly;
        glBegin (GL_TRIANGLES);
        //fprintf(stderr, "# slices: %f\n", (z + d - zstart)/thickness);
        for (float k = zstart; k <= z + d; k += thickness)
        {
                float ep1[4] = { x, y, k, 1 };
                float ep2[4] = { x + w, y, k, 1 };
                float ep3[4] = { x + w, y + h, k, 1 };
                float ep4[4] = { x, y + h, k, 1 };
                
                poly.reset ();
                poly.appendVert (ep1);
                poly.appendVert (ep2);
                poly.appendVert (ep3);
                poly.appendVert (ep4);
                
                // Clip to bounding box
                for (unsigned int n = 0; n < clipPlanes.size (); n++) {
                        poly.clip (clipPlanes[n]);
                }
                emitTriangles (poly);
        } 
        glEnd ();
} 

void EyeSlicedCube::render () 
{
        
        //glPushAttrib(GL_POLYGON_BIT | GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT); 
        // Transform each plane, and figure out whether the eye is clipped.
        // This determines whether the plane should be drawn before
        // or after the volume
        vector < int >clipPlanesBefore, clipPlanesAfter;
        matrix4_t modelview = GetModelView ();
        if (renderClipPlanes)
        {
                for (unsigned int i = 0; i < userClipPlanes.size (); i++)
                {
                        Plane p = userClipPlanes[i];
                        float eye[4] = { 0, 0, 0, 1 };
                        p.transform (modelview);
                        if (p.retains (eye))
                                clipPlanesAfter.push_back (i);
                        
                        else
                                clipPlanesBefore.push_back (i);
                }
        }
        for (unsigned int j = 0; j < clipPlanesBefore.size (); j++)
        {
                drawClipPlane (userClipPlanes[clipPlanesBefore[j]]);
        } 
        renderViewSliced (modelview);
        for (unsigned int k = 0; k < clipPlanesAfter.size (); k++)
        {
                drawClipPlane (userClipPlanes[clipPlanesAfter[k]]);
        } 
} 


#define LARGE_VAL 10000
static inline void CalcPointHelper (int i, int j, const float p[3],
                                    const float t[3], const float b[3],
                                    float polyPoint[4]) 
{
        polyPoint[0] = p[0] + i * LARGE_VAL * t[0] + j * LARGE_VAL * b[0];
        polyPoint[1] = p[1] + i * LARGE_VAL * t[1] + j * LARGE_VAL * b[1];
        polyPoint[2] = p[2] + i * LARGE_VAL * t[2] + j * LARGE_VAL * b[2];
        polyPoint[3] = 1;
} 

void EyeSlicedCube::drawClipPlane (const Plane & plane) 
{                              
        ConvexPolygon poly;
        {
                float p[3], n[3], t[3], b[3], polyPoint[4];
                plane.getNormal (n);
                plane.getPoint (p);
                if (n[0] != 0 || n[1] != 0)
                {
                        t[0] = -n[1];
                        t[1] = n[0];
                        t[2] = 0;
                        b[0] = -n[0] * n[2];
                        b[1] = -n[1] * n[2];
                        b[2] = n[0] * n[0] + n[1] * n[1];
                }
                
                else
                {
                        t[1] = t[2] = 0;
                        t[0] = 1;
                        b[1] = b[2] = 0;
                        b[1] = 1;
                }
                CalcPointHelper (-1, -1, p, t, b, polyPoint);
                poly.appendVert (polyPoint);
                CalcPointHelper (-1, 1, p, t, b, polyPoint);
                poly.appendVert (polyPoint);
                CalcPointHelper (1, 1, p, t, b, polyPoint);
                poly.appendVert (polyPoint);
                CalcPointHelper (1, -1, p, t, b, polyPoint);
                poly.appendVert (polyPoint);
        }
        
        vector < Plane > clipPlanes;
        clipPlanes.push_back (Plane (1, 0, 0, -xmin));     // x >= xmin
        clipPlanes.push_back (Plane (-1, 0, 0, xmax));      // x <= xmax
        clipPlanes.push_back (Plane (0, 1, 0, -ymin));
        clipPlanes.push_back (Plane (0, -1, 0, ymax));
        clipPlanes.push_back (Plane (0, 0, 1, -zmin));
        clipPlanes.push_back (Plane (0, 0, -1, zmax));
        
        // Clip to bounding box
        for (unsigned int n = 0; n < clipPlanes.size (); n++)
        {
                poly.clip (clipPlanes[n]);
        } 
        float norm[3];
        plane.getNormal (norm);
        
        
        // Added so the clip plane appears in front of the 
        // texture slices always, even though there may be 
        // depth calculation error
        glPushAttrib (GL_ENABLE_BIT);
        {
                glDisable (GL_DEPTH_TEST);
                glBegin (GL_POLYGON);
                {
                        glNormal3fv (norm);
                        emitPolygon (poly);
                } 
                glEnd ();
        } 
        glPopAttrib ();
} 

void EyeSlicedCube::getPointOfInterest (float p[4]) 
{
        p[0] = p[1] = p[2] = 0;
        p[3] = 1;
} 

//*************
// PLANE CLASS
//*************
Plane::Plane (myfloat a, myfloat b, myfloat c, myfloat d) 
{
        this->a = a;
        this->b = b;
        this->c = c;
        this->d = d;
        assert (a != 0 || b != 0 || c != 0);
} 

Plane::Plane (myfloat coeffs[4]) 
{
        a = coeffs[0];
        b = coeffs[1];
        c = coeffs[2];
        d = coeffs[3];
        assert (a != 0 || b != 0 || c != 0);
} 

void Plane::getNormal (myfloat norm[3]) const
{
        norm[0] = a;
        norm[1] = b;
        norm[2] = c;
        
        // should we normalize?
} 

static void fatal (const char *str) 
{
        cerr << str << flush;
        abort ();
} 

void Plane::getPoint (myfloat point[3]) const 
{
        if (a != 0)
        {
                point[0] = -d / a;
                point[1] = point[2] = 0;
        }
        
        else if (b != 0)
        {
                point[1] = -d / b;
                point[0] = point[2] = 0;
        }
        
        else if (c != 0)
        {
                point[2] = -d / c;
                point[1] = point[0] = 0;
        }
        
        else
        {
                fatal ("Internal error\n");
        }
}

void Plane::getCoeffs (myfloat coeffs[4]) const
{
        coeffs[0] = a;
        coeffs[1] = b;
        coeffs[2] = c;
        coeffs[3] = d;
} 

static void matrix_postmultiply_vec_4 (float *p, matrix4_t * l,
                                       float *r) 
{
        assert (p != r);
        p[0] = l->m00 * r[0] + l->m10 * r[1] + l->m20 * r[2] + l->m30 * r[3];
        p[1] = l->m01 * r[0] + l->m11 * r[1] + l->m21 * r[2] + l->m31 * r[3];
        p[2] = l->m02 * r[0] + l->m12 * r[1] + l->m22 * r[2] + l->m32 * r[3];
        p[3] = l->m03 * r[0] + l->m13 * r[1] + l->m23 * r[2] + l->m33 * r[3];
} 

void Plane::transform (matrix4_t & t) 
{
        matrix4_t invt;
        matrix_invert_4 (&invt, &t);
        float coeffs[4] = { a, b, c, d };
        float newcoeffs[4];
        matrix_postmultiply_vec_4 (newcoeffs, &invt, coeffs);
        a = newcoeffs[0];
        b = newcoeffs[1];
        c = newcoeffs[2];
        d = newcoeffs[3];
} 

bool Plane::retains (float point[3]) const
{
        return (a * point[0] + b * point[1] + c * point[2]) >= d;
}


//**********************
// CONVEX POLYGON CLASS
//**********************
ConvexPolygon::ConvexPolygon (const VertexVector & vertices) 
{
        this->vertices.insert (this->vertices.begin (), vertices.begin (),
                               vertices.end ());
} 

void ConvexPolygon::appendVert (const float p[4]) 
{
        Point point (p[0], p[1], p[2], p[3]);
        vertices.push_back (point);
} 

void ConvexPolygon::insertVert (int index, const float p[4]) 
{
        assert ((unsigned) index < vertices.size ());
        Point point (p[0], p[1], p[2], p[3]);
        vertices.insert (vertices.begin () + index, point);
} 

static bool isRetained (const float point[4], const float plane[4]) 
{
        return (point[0] * plane[0] + point[1] * plane[1] + point[2] * plane[2] +
                point[3] * plane[3]) >= 0;
}

static bool isRetained (const Point & p, const float plane[4]) 
{
        return isRetained ((const float *) &p, plane);
} 

// Finds the intersection of the line through 
// p1 and p2 with the given plane, putting the
// result in intersect.
//
// If the line lies in the plane, an arbitrary 
// point on the line is returned.
//
// Simple equation, checked it up on 
// http://astronomy.swin.edu.au/pbourke/geometry/planeline/
static bool findIntersection (const float p1[4], const float p2[4],
                              const float plane[4], float intersect[4]) 
{
        float a = plane[0];
        float b = plane[1];
        float c = plane[2];
        float d = plane[3];
        float x1 = p1[0] / p1[3];
        float y1 = p1[1] / p1[3];
        float z1 = p1[2] / p1[3];
        float x2 = p2[0] / p2[3];
        float y2 = p2[1] / p2[3];
        float z2 = p2[2] / p2[3];
        float uDenom = a * (x1 - x2) + b * (y1 - y2) + c * (z1 - z2);
        float uNumer = a * x1 + b * y1 + c * z1 + d;
        if (uDenom == 0)
        {        
                // plane parallel to line
                cout << "Unexpected code path: clip.cpp\n" << endl;
                if (uNumer == 0)
                {
                        for (int i = 0; i < 4; i++)
                                intersect[i] = p1[i];
                        return true;
                }
                return false;
        }
        float u = uNumer / uDenom;
        intersect[0] = x1 + u * (x2 - x1);
        intersect[1] = y1 + u * (y2 - y1);
        intersect[2] = z1 + u * (z2 - z1);
        intersect[3] = 1;
        
        return true;
}

static bool findIntersection (const Point & p1, const Point & p2,
                              const float plane[4], float intersect[4]) 
{
        return findIntersection ((const float *) &p1, 
                                 (const float *) &p2, plane, intersect);
} 

void ConvexPolygon::clip (const Plane & clipPlane) 
{
        if (vertices.size () == 0)
                return;
        VertexVector clippedVerts;
        clippedVerts.reserve (2 * vertices.size ());
        
        // The algorithm is as follows: for each vertex
        // in the current poly, check to see which 
        // half space it is in: clipped or retained.
        // If the previous vertex was in the other
        // half space, find the intersect of that edge
        // with the clip plane and add that intersect
        // point to the new list of vertices.  If the 
        // current vertex is in the retained half-space,
        // add it to the new list as well.
        float plane[4], intersect[4];
        clipPlane.getCoeffs (plane);
        bool prevRetained = isRetained (vertices[0], plane);
        if (prevRetained)
                clippedVerts.push_back (Point (vertices[0]));
        for (unsigned int i = 1; i < vertices.size (); i++)
        {
                bool retained = isRetained (vertices[i], plane);
                if (retained != prevRetained)
                {
                        bool found =
                                findIntersection (vertices[i - 1], 
                                                  vertices[i], plane,
                                                  intersect);
                        if (!found)
                                fatal ("Internal error\n");
                        clippedVerts.push_back (Point (intersect));
                }
                if (retained)
                {
                        clippedVerts.push_back (Point (vertices[i]));
                }
                prevRetained = retained;
        }
        bool retained = isRetained (vertices[0], plane);
        if (retained != prevRetained)
        {
                bool found =
                        findIntersection (vertices[vertices.size () - 1], 
                                          vertices[0],
                                          plane, intersect);
                if (!found)
                        fatal ("Internal error\n");
                clippedVerts.push_back (Point (intersect));
        }
        vertices.clear ();
        vertices.insert (vertices.begin (), clippedVerts.begin (),
                         clippedVerts.end ());
}

void Point::print () const
{
        cout << '(' << x << ','  <<y << ',' << z << ','  <<w << ')' << endl;
} 
