/*Copyright (c) 2004, Stanford University
  All rights reserved.*/
 
// This file contains the classes that are necessary to
// render slices through a volume cuboid to resample
// it for volume rendering.  
//
// The EyeSlicedCube provide the main functionality for 
// this, and is implemented using the ConvexPolygon class,
// that provides methods that clip polygons against 
// half-planes.  

#ifndef _CLIP_H
#define _CLIP_H

#include <vector>
#include <iostream>
#ifdef WIN_32
#include "ATIExtensions.h"
#endif

extern "C" { 
#include "vec.h"
}

class Plane;
typedef struct Point{
        float x;
        float y;
        float z;
        float w;
        Point() {
                x = y = z = 0;
                w = 1;
	}
        
	Point(float x, float y, float z) {
		this->x = x; 
		this->y = y; 
		this->z = z;
		this->w = 1;
	}
        
	Point(float x, float y, float z, float w) {
		this->x = x; 
		this->y = y; 
		this->z = z;
		this->w = w;
	}
        
	Point(const float p[4]) {
		x = p[0]; y = p[1]; z = p[2]; w = p[3]; 
	}
        
	const Point &operator=(const Point &p) {
		this->x = p.x;
		this->y = p.y; 
		this->z = p.z;
		this->w = p.w;
		return *this;
	}
        
	float X() { return x / w; }
	float Y() { return y / w; }
	float Z() { return z / w; }
	void print() const;
} Point;

typedef std::vector<Point> VertexVector;
typedef std::vector<Point> PointList;


typedef struct BBox {
        float x1, y1, z1, x2, y2, z2; 
        BBox(float x1, float y1, float z1, 
             float x2, float y2, float z2) {
                this->x1 = x1;  this->y1 = y1;  this->z1 = z1;
                this->x2 = x2;  this->y2 = y2;  this->z2 = z2;
        }
        void print() {
                std::cout << "(" << x1 << ", " << y1 << ", "  << z1 << ", "
                          << x2 << ", "  << y2 << ", " << z2 << ")" 
                          << std::endl;
        }
} BBox;

// *************************
// POINT OF INTEREST CLASSES
// *************************

// An abstract interface for an object that can register
// interest with a PointOfInterest for updates when the POI
// changes.
class POIChangeListener {
public:
        virtual void poiChanged(float x, float y, float z) = 0;
};


// An abstract interface for an object that acts as a PointOfInterest.
// Cameras (see camera.h) can be automatically focused on such objects.
class PointOfInterest {
public:
        virtual ~PointOfInterest(){};
	virtual void getPointOfInterest(float p[4]) = 0;
	
	// Convenience: returns a free-store float[4] 
	float *getPointOfInterest() {
                float *copy = (float *)malloc(4 * sizeof(float));
                getPointOfInterest(copy);
                return copy;
	}
	void addListener(POIChangeListener *l);
	void removeListener(POIChangeListener *l);
protected:
	void dispatchNewPOI(float x, float y, float z);
	std::vector<POIChangeListener *> poiListeners;
};

//********************
// BOUNDING BOX CLASS
//********************


// An immutable bounding box.
class BoundingBox {
public:
        BoundingBox() {
                float box[6] = { 0,0,0,0,0,0 };
                set(box);
        }
        BoundingBox(float x, float y, float z, float w, float h, float d) {
                float box[6] = { x, y, z, w, h, d };
                set(box);
        }
        
        BoundingBox(float box[6]) {
                set(box);
        }
        
        // Returns (x,y,z,w,h,d);
        void get(float box[6]);
        void set(float b[6]) {
                for (int i = 0; i < 6; i++) box[i] = b[i];
        }
        
        float getX() { return box[0]; }
        float getY() { return box[1]; }
        float getZ() { return box[2]; }
        float getWidth() { return box[3]; }
        float getHeight() { return box[4]; }
        float getDepth() { return box[5]; }
private:
        float box[6];
};


// Renders a cube with obj coords [0, 1]^3, centered at origin. 
// The cube is sliced by planes perpendicular to the view direction, 
// in back to front order.  
class EyeSlicedCube : public PointOfInterest {
public:
        EyeSlicedCube(float xmin, float xmax,
                      float ymin, float ymax,
                      float zmin, float zmax,
                      float SliceThickness = 0.01f);
        
        EyeSlicedCube(float SliceThickness = 0.01f);
        
        void render();
        void getPointOfInterest(float p[4]);
        void setSliceThickness(float t);
        float getSliceThickness();
        
        void addClipPlane(const Plane &clipPlane);
        void removeClipPlane(int index);
        const Plane &getClipPlane(int index);
        
        bool isRenderingClipPlanes() { return renderClipPlanes; }
        void setRenderingClipPlanes(bool b) { renderClipPlanes = b; }
        
        BBox getObjSpaceBBox() { return BBox(xmin, ymin, zmin, xmax, ymax, zmax); }
        
protected:
        void renderViewSliced(matrix4_t modelview);
        
        void drawClipPlane(const Plane &plane);
        
        float xmin, xmax, ymin, ymax, zmin, zmax;
        float thickness;
        std::vector<Plane> userClipPlanes;
        bool renderClipPlanes;
	PointList cubeVertices;	
        bool initialized;
};



// This clas provides the functionality needed to clip polygons against 
//
// This is used in the volume rendering demo to clip view-aligned 
// slicing planes against the bounds of the object volume.  
// In addition, "user-defined" clipping planes are easily included
// to trim away parts of the sliced volume.
//
// Does not check that constructed Polygon is convex (corner-cut).
// Allows clipping by a plane, where ax + by + cz + d >= 0 is retained.
class ConvexPolygon {
public:
        // Constructing a poly
        ConvexPolygon(const VertexVector &vertices);
        ConvexPolygon() { }
        void appendVert(const float p[4]);
        void insertVert(int index, const float p[4]);
        
        const VertexVector &verts() const { return vertices; }
        
        // Clips the polygon, retaining the portion where ax + by + cz + d >= 0
        void clip(const Plane &clipPlane);
        void reset() { vertices.erase(vertices.begin(), vertices.end()); };
private:
        VertexVector vertices;
};



// An abstraction for a plane.  The ConvexPolygon clips away regions
// of the plane where ax + by + cz + d < 0.
class Plane {
public:
        Plane(float a, float b, float c, float d);
        Plane(float coeffs[4]);
        void getNormal(float norm[3]) const;
        void getPoint(float point[3]) const;
        void getCoeffs(float coeffs[4]) const;
        void transform(matrix4_t &t);
        bool retains(float point[3]) const;
        bool clips(float point[3]) const { return !retains(point); }
private:
        float a, b, c, d;  // equation ax + by + cz + d = 0
};

#endif
