/* Copyright (c) 2004, Stanford University
   All rights reserved.*/

#include "subvolume.h"
    
// Returns pow of 2 that is greater than or equal to the given val
static int Power2GreaterEq (int val) 
{
    if (val == 0)
        return val;
    int ret = 1;
    while (ret < val)
        ret *= 2;
    return ret;
}

SubvolumeManager::SubvolumeManager (int xres, int yres, int zres,
                                    float voxelWidth, float voxelHeight,
                                    float voxelDepth, int x, int y, int z,
                                    int w, int h, int d, int overlap,
                                    bool nomessage)
{
    width  = xres;
    height = yres;
    depth  = zres;
    voxOverlap = overlap*2;
    this->voxWidth  = voxelWidth;
    this->voxHeight = voxelHeight;
    this->voxDepth  = voxelDepth;
    maxDim = (width * voxWidth) > (height * voxHeight) ? 
			(width * voxWidth) : (height * voxHeight);
    maxDim = maxDim > (depth * voxDepth) ? maxDim : (depth * voxDepth);
    
    renderX = x*voxelWidth;
    renderY = y*voxelHeight;
    renderZ = z*voxelDepth;
    renderWidth  = (w-voxOverlap)*voxelWidth;
    renderHeight = (h-voxOverlap)*voxelHeight;
    renderDepth  = (d-voxOverlap)*voxelDepth;
    
    loadedX = (x == 0) ? x : x - 1;
    loadedY = (y == 0) ? y : y - 1;
    loadedZ = (z == 0) ? z : z - 1;
    loadedWidth  = (w-voxOverlap) + (x - loadedX) + ((x + (w-voxOverlap) == xres) ? 0 : 1);
    loadedHeight = (h-voxOverlap) + (y - loadedY) + ((y + (h-voxOverlap) == yres) ? 0 : 1);
    loadedDepth  = (d-voxOverlap) + (z - loadedZ) + ((z + (d-voxOverlap) == zres) ? 0 : 1);
    slicesPerVoxel = 1.0;
    int glWidth  = Power2GreaterEq (loadedWidth);
    int glHeight = Power2GreaterEq (loadedHeight);
    int glDepth  = Power2GreaterEq (loadedDepth);
    
    float fractionUsed = (float) ((w-voxOverlap) * (h-voxOverlap) * (d-voxOverlap)) / (glWidth * glHeight * glDepth);
    if (!nomessage)
        fprintf (stderr, "%sRendered portion is %d%c of loaded GL texture\n", 
                  (fractionUsed < 0.5) ? "WARNING: " : "", 
                  (int) (fractionUsed * 100), '%');
} 

EyeSlicedCube * SubvolumeManager::getResamplingCube ()
{
     float w = (float)renderWidth / maxDim;
     float h = (float)renderHeight/ maxDim;
     float d = (float)renderDepth / maxDim; 
     float x = (float)renderX / maxDim - ((width-2.0f) *voxWidth /maxDim)/2.0f;
     float y = (float)renderY / maxDim - ((height-2.0f)*voxHeight/maxDim)/2.0f;
     float z = (float)renderZ / maxDim - ((depth-2.0f) *voxDepth /maxDim)/2.0f;
		 
     //We want to sample to the smallest voxel dimension
     float minVoxSide = voxWidth < voxHeight ? voxWidth:voxHeight;
     minVoxSide = minVoxSide < voxDepth ? minVoxSide:voxDepth;
     
     return new EyeSlicedCube (x, x + w, y, y + h, z, z + d,
                               1.0f / (maxDim/minVoxSide * slicesPerVoxel));
}


// This is a little tricky.  The geometry for the object is emitted in
// a subpart of [0-1]^3 object space. [0-1]^3 itself represents the
// texture mapping for the entire GL 3D texture.  The loaded subvolume
// is located near the origin of the GL texture, but may not fill the
// entire GL texture (zero padding is used if it doesn't).  Further,
// the rendered subvolume may not fill the loaded subvolume, because
// the loaded subvolume may need to be larger to prevent boundary
// artifacts due to incorrect texture filtering.
//
// Thus, to transform from obj coordinates to the correct texture
// coordinates for the textures returned by this class, we need to:
// 1. Transform from emitted object coordinates to [0-1]^3,
// 2. Transform from [0-1]^3 to the sub-part of the texture
matrix4_t SubvolumeManager::getObjToTexTransform ()
{
    matrix4_t temp1, temp2;
    
    // Obj to normalized ([0-1]^3) box
    matrix4_t objToNormalizedBox; {
        matrix_translate_4(&temp1,
                           -((float)renderX / maxDim - (width *voxWidth  / maxDim)/2.0f), 
                           -((float)renderY / maxDim - (height*voxHeight / maxDim)/2.0f), 
                           -((float)renderZ / maxDim - (depth *voxDepth  / maxDim)/2.0f));
        matrix_scale_4(&temp2,
                       (float)maxDim / (renderWidth +voxOverlap), 
                       (float)maxDim / (renderHeight+voxOverlap), 
                       (float)maxDim / (renderDepth +voxOverlap));
        // scale first, then translate
        matrix_multiply_4(&objToNormalizedBox,
                          &temp2, &temp1);
    } 
    
    // [0,1]^3 to data for rendered subvolume
    matrix4_t normToRenderSubvolume;
    {
        float widthGLTex  = (float)Power2GreaterEq(getLoadedWidth ());
        float heightGLTex = (float)Power2GreaterEq(getLoadedHeight ());
        float depthGLTex  = (float)Power2GreaterEq(getLoadedDepth ());
        matrix_translate_4(&temp1, 
                           (getRenderX()/voxWidth  - getLoadedX ()) / widthGLTex, 
                           (getRenderY()/voxHeight - getLoadedY ()) / heightGLTex, 
                           (getRenderZ()/voxDepth  - getLoadedZ ()) / depthGLTex);
        matrix_scale_4(&temp2, 
                       getRenderWidth() /voxWidth  / widthGLTex,
                       getRenderHeight()/voxHeight / heightGLTex,
                       getRenderDepth() /voxDepth  / depthGLTex);
        matrix_multiply_4(&normToRenderSubvolume, &temp1, &temp2);
    } 
    matrix4_t ret;
    matrix_multiply_4 (&ret, &normToRenderSubvolume, &objToNormalizedBox);
    return ret;
}


