/* Copyright (c) 2004, Stanford University
   All rights reserved.*/

#ifndef _SUBVOLUME_H
#define _SUBVOLUME_H

#include <map>
#include "clip.h"
#include <string>

extern "C" {
#include "vec.h"
}

class SubvolumeManager {
public:
    // Manages all the textures, transforms, etc, to render
    // the subvol with lower left corner (x,y,z) and dimensions
    // (w,h,d).  The original voxel data is a 1-channel 
    // density file with the given dimensions (width, height, depth).
    //
    // If nomessage is passed as true, then no messages are sent
    // to the console regarding the efficiency of the bound
    // versus used GL 3D textures.
    SubvolumeManager(
        int xres, int yres, int zres, 
        float voxelWidth, 
        float voxelHeight, 
        float voxelDepth,
        int x, int y, int z,
        int w, int h, int d,
        int overlap,
        bool nomessage = false);
    
    //***********
    // PROPERTIES
    //***********
    
    int getVolumeFileWidth() { return width; }
    int getVolumeFileHeight() { return height; }
    int getVolumeFileDepth() { return depth; }
    int getVoxelOverlap() { return voxOverlap; }
    
    // The location and dimensions of the subvolume to be rendered,
    // relative to the entire volume represented by the data file.
    float getRenderX() { return renderX; }
    float getRenderY() { return renderY; }
    float getRenderZ() { return renderZ; }
    float getRenderWidth()  { return renderWidth; }
    float getRenderHeight() { return renderHeight; }
    float getRenderDepth()  { return renderDepth; }
    
    float getVoxelWidth()  { return voxWidth; }
    float getVoxelHeight() { return voxHeight; }
    float getVoxelDepth()  { return voxDepth; }
    
    float getTotalWidth()  { return voxWidth * width; }
    float getTotalHeight() { return voxHeight * height; }
    float getTotalDepth()  { return voxDepth * depth; }
    
    //***************
    // HELPER METHODS
    //***************
    
    // These are measured relative to the z-axis.  Thus, 2 slices
    // per voxel means that the slices will be sufficiently close
    // that when viewed along the z-axis, each voxel will be sliced
    // by 2 resampling planes.  
    float getNumSlicesPerVoxel() { return slicesPerVoxel; }
    void setNumSlicesPerVoxel(float s) { slicesPerVoxel = s; }
    
    
    // Returns an EyeSlicedCube that will emit slices of geometry
    // at the specified separation (specified by setNumSlicesPerVoxel)
    // to resample the portion of [0,1]^3 3D space that would contain
    // the subvolume, assuming that the entire volume is contained
    // in [0,1]^3.
    EyeSlicedCube *getResamplingCube();
    
    // Returns a transformation that translates from the 
    // coordinates of the rendered subvolume to the texture 
    // coordinates within the GL textures encapsulated by 
    // the getXXXTtexture() methods.  
    // This matrix can be concatenated to the left of a
    // WorldToObject transformation, to obtain a matrix 
    // that translates from world space to the appropriate
    // texture coordinates.
    // The matrix returned by this method encodes information
    // about the location of the rendered subvolume within
    // the loaded subvolume, as well as information about 
    // the location of the loaded subvolume within the 
    // bound GL 3D texture (the loaded subvolume may not 
    // have power of 2 dimensions).
    matrix4_t getObjToTexTransform();
    
    
    // The location and dimensions of the subvolume that was
    // loaded, relative to the total volume reprenseted by the 
    // data file.
    int getLoadedX() { return loadedX; }
    int getLoadedY() { return loadedY; }
    int getLoadedZ() { return loadedZ; }
    int getLoadedWidth() { return loadedWidth; }
    int getLoadedHeight() { return loadedHeight; }
    int getLoadedDepth() { return loadedDepth; }
    
    
private:    
    //*************
    // PRIVATE DATA
    //*************
    
    int width, height, depth;
    float maxDim;
    float voxWidth, voxHeight, voxDepth;
    int loadedX, loadedY, loadedZ;
    int loadedWidth, loadedHeight, loadedDepth;
    float renderX, renderY, renderZ;
    float renderWidth, renderHeight, renderDepth;
    int voxOverlap;
    float slicesPerVoxel;
};

#endif
