/* Copyright (c) 2004, Stanford University
   All rights reserved.*/

/*
 *  vec.c
 *
 *  Matrix and vector math routines.
 *
 *  Kekoa Proudfoot
 *  12/20/99
 */

#include <assert.h>
#include <math.h>

#include "vec.h"

#define DEG_TO_RAD_MULT 0.0174532925f

void
vec_normalize_3 (float *p, float *v)
{
        float scale = 1.0f / (float) sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
        p[0] = v[0] * scale;
        p[1] = v[1] * scale;
        p[2] = v[2] * scale;
}

void
matrix_identity_3 (matrix3_t *p)
{
        p->m00 = 1.0f; p->m01 = 0.0f; p->m02 = 0.0f;
        p->m10 = 0.0f; p->m11 = 1.0f; p->m12 = 0.0f;
        p->m20 = 0.0f; p->m21 = 0.0f; p->m22 = 1.0f;
}

void
matrix_invert_3 (matrix3_t *p, matrix3_t *m)
{
        float m00 = m->m00, m01 = m->m01, m02 = m->m02;
        float m10 = m->m10, m11 = m->m11, m12 = m->m12;
        float m20 = m->m20, m21 = m->m21, m22 = m->m22;
        
        float det0, inv0;
        
#define det22(a1,a2,b1,b2) (a1*b2-a2*b1)
        
        float cof00 = det22(m11, m12, m21, m22);
        float cof01 = det22(m12, m10, m22, m20);
        float cof02 = det22(m10, m11, m20, m21);
        
        float cof10 = det22(m21, m22, m01, m02);
        float cof11 = det22(m22, m20, m02, m00);
        float cof12 = det22(m20, m21, m00, m01);
        
        float cof20 = det22(m01, m02, m11, m12);
        float cof21 = det22(m02, m00, m12, m10);
        float cof22 = det22(m00, m01, m10, m11);
        
#undef det22        

        det0  = m00 * cof00 + m01 * cof01 + m02 * cof02;
        
        inv0  = (det0 == 0.0f) ? 0.0f : 1.0f / det0;
        
        p->m00 = cof00 * inv0;
        p->m10 = cof01 * inv0;
        p->m20 = cof02 * inv0;
        
        p->m01 = cof10 * inv0;
        p->m11 = cof11 * inv0;
        p->m21 = cof12 * inv0;
        
        p->m02 = cof20 * inv0;
        p->m12 = cof21 * inv0;
        p->m22 = cof22 * inv0;
}

void
matrix_multiply_3 (matrix3_t *p, matrix3_t *l, matrix3_t *r)
{
        assert(p != l);
        assert(p != r);
        
#define prod(i,j,n) (l->m##i##n * r->m##n##j)
#define dot3(i,j) p->m##i##j = prod(i,j,0)+prod(i,j,1)+prod(i,j,2)
        
        dot3(0,0); dot3(0,1); dot3(0,2);
        dot3(1,0); dot3(1,1); dot3(1,2);
        dot3(2,0); dot3(2,1); dot3(2,2);
        
#undef prod
#undef dot3
}

void
matrix_multiply_vec_3 (float *p, matrix3_t *l, float *r)
{
        assert (p != r);
        p[0] = l->m00 * r[0] + l->m01 * r[1] + l->m02 * r[2];
        p[1] = l->m10 * r[0] + l->m11 * r[1] + l->m12 * r[2];
        p[2] = l->m20 * r[0] + l->m21 * r[1] + l->m22 * r[2];
}

void
matrix_rotate_3 (matrix3_t *p, float angle, float x, float y, float z)
{
        float len = (float) sqrt(x * x + y * y + z * z);
        float inv = (len == 0.0f) ? 0.0f : 1.0f / len;
        
        float theta = angle * DEG_TO_RAD_MULT;
        float s = (float) sin(theta);
        float c = (float) cos(theta);
        
        x *= inv;
        y *= inv;
        z *= inv;
        
        p->m00 = x * x * (1.0f - c) + c;
        p->m01 = y * x * (1.0f - c) - z * s;
        p->m02 = z * x * (1.0f - c) + y * s;
        
        p->m10 = x * y * (1.0f - c) + z * s;
        p->m11 = y * y * (1.0f - c) + c;
        p->m12 = z * y * (1.0f - c) - x * s;
        
        p->m20 = x * z * (1.0f - c) - y * s;
        p->m21 = y * z * (1.0f - c) + x * s;
        p->m22 = z * z * (1.0f - c) + c;
}

void
matrix_scale_3 (matrix3_t *p, float x, float y, float z)
{
        p->m00 = x; p->m01 = 0.0f; p->m02 = 0.0f;
        p->m10 = 0.0f; p->m11 = y; p->m12 = 0.0f;
        p->m20 = 0.0f; p->m21 = 0.0f; p->m22 = z;
}

void
matrix_transpose_3 (matrix3_t *p, matrix3_t *m)
{
        assert(p != m);
        p->m00 = m->m00; p->m01 = m->m10; p->m02 = m->m20;
        p->m10 = m->m01; p->m11 = m->m11; p->m12 = m->m21;
        p->m20 = m->m02; p->m21 = m->m12; p->m22 = m->m22;
}

void
matrix_affine_4 (matrix3_t *p, matrix4_t *m)
{
        p->m00 = m->m00; p->m01 = m->m01; p->m02 = m->m02;
        p->m10 = m->m10; p->m11 = m->m11; p->m12 = m->m12;
        p->m20 = m->m20; p->m21 = m->m21; p->m22 = m->m22;
}

void
matrix_frustum_4 (matrix4_t *p, float l, float r, float b, float t, float n,
		  float f)
{
        float invdx = (l == r) ? 0.0f : 1.0f / (r - l);
        float invdy = (b == t) ? 0.0f : 1.0f / (t - b);
        float invdz = (n == f) ? 0.0f : 1.0f / (n - f);
        float twon = n + n;
        
        p->m00 = twon * invdx;
        p->m01 = 0.0f;
        p->m02 = (r + l) * invdx;
        p->m03 = 0.0f;
        
        p->m10 = 0.0f;
        p->m11 = twon * invdy;
        p->m12 = (t + b) * invdy;
        p->m13 = 0.0f;
        
        p->m20 = 0.0f;
        p->m21 = 0.0f;
        p->m22 = (f + n) * invdz;
        p->m23 = twon * f * invdz;
        
        p->m30 = 0.0f;
        p->m31 = 0.0f;
        p->m32 = -1.0f;
        p->m33 = 0.0f;
}

void
matrix_identity_4 (matrix4_t *p)
{
        p->m00 = 1.0f; p->m01 = 0.0f; p->m02 = 0.0f; p->m03 = 0.0f;
        p->m10 = 0.0f; p->m11 = 1.0f; p->m12 = 0.0f; p->m13 = 0.0f;
        p->m20 = 0.0f; p->m21 = 0.0f; p->m22 = 1.0f; p->m23 = 0.0f;
        p->m30 = 0.0f; p->m31 = 0.0f; p->m32 = 0.0f; p->m33 = 1.0f;
}

void
matrix_invert_4 (matrix4_t *p, matrix4_t *m)
{
        float m00 = m->m00, m01 = m->m01, m02 = m->m02, m03 = m->m03;
        float m10 = m->m10, m11 = m->m11, m12 = m->m12, m13 = m->m13;
        float m20 = m->m20, m21 = m->m21, m22 = m->m22, m23 = m->m23;
        float m30 = m->m30, m31 = m->m31, m32 = m->m32, m33 = m->m33;
        
        float det0, inv0;
        
#define det33(a1,a2,a3,b1,b2,b3,c1,c2,c3) \
    (a1*(b2*c3-b3*c2) + b1*(c2*a3-a2*c3) + c1*(a2*b3-a3*b2))
        
        float cof00 =  det33(m11, m12, m13, m21, m22, m23, m31, m32, m33);
        float cof01 = -det33(m12, m13, m10, m22, m23, m20, m32, m33, m30);
        float cof02 =  det33(m13, m10, m11, m23, m20, m21, m33, m30, m31);
        float cof03 = -det33(m10, m11, m12, m20, m21, m22, m30, m31, m32);
        
        float cof10 = -det33(m21, m22, m23, m31, m32, m33, m01, m02, m03);
        float cof11 =  det33(m22, m23, m20, m32, m33, m30, m02, m03, m00);
        float cof12 = -det33(m23, m20, m21, m33, m30, m31, m03, m00, m01);
        float cof13 =  det33(m20, m21, m22, m30, m31, m32, m00, m01, m02);
        
        float cof20 =  det33(m31, m32, m33, m01, m02, m03, m11, m12, m13);
        float cof21 = -det33(m32, m33, m30, m02, m03, m00, m12, m13, m10);
        float cof22 =  det33(m33, m30, m31, m03, m00, m01, m13, m10, m11);
        float cof23 = -det33(m30, m31, m32, m00, m01, m02, m10, m11, m12);
        
        float cof30 = -det33(m01, m02, m03, m11, m12, m13, m21, m22, m23);
        float cof31 =  det33(m02, m03, m00, m12, m13, m10, m22, m23, m20);
        float cof32 = -det33(m03, m00, m01, m13, m10, m11, m23, m20, m21);
        float cof33 =  det33(m00, m01, m02, m10, m11, m12, m20, m21, m22);
        
#undef det33
        
        
        det0  = m30 * cof30 + m31 * cof31 + m32 * cof32 + m33 * cof33;
        inv0  = (det0 == 0.0f) ? 0.0f : 1.0f / det0;
        
        p->m00 = cof00 * inv0;
        p->m10 = cof01 * inv0;
        p->m20 = cof02 * inv0;
        p->m30 = cof03 * inv0;
        
        p->m01 = cof10 * inv0;
        p->m11 = cof11 * inv0;
        p->m21 = cof12 * inv0;
        p->m31 = cof13 * inv0;
        
        p->m02 = cof20 * inv0;
        p->m12 = cof21 * inv0;
        p->m22 = cof22 * inv0;
        p->m32 = cof23 * inv0;
        
        p->m03 = cof30 * inv0;
        p->m13 = cof31 * inv0;
        p->m23 = cof32 * inv0;
        p->m33 = cof33 * inv0;
}

void
matrix_lookat_4 (matrix4_t *p, float ex, float ey, float ez, float cx,
		 float cy, float cz, float ux, float uy, float uz)
{
        float len;
        float inv;
        
#define length(x,y,z) ((float) sqrt(x * x + y * y + z * z))
#define norm(r) \
    len = length(p->m##r##0, p->m##r##1, p->m##r##2); \
    inv = (len == 0.0f) ? 0.0f : 1.0f / len; \
    p->m##r##0 *= inv; p->m##r##1 *= inv; p->m##r##2 *= inv
        
        p->m20 = ex - cx;
        p->m21 = ey - cy;
        p->m22 = ez - cz;
        norm(2);
        
        p->m00 = uy * p->m22 - uz * p->m21;
        p->m01 = uz * p->m20 - ux * p->m22;
        p->m02 = ux * p->m21 - uy * p->m20;
        norm(0);
        
        p->m10 = p->m21 * p->m02 - p->m22 * p->m01;
        p->m11 = p->m22 * p->m00 - p->m20 * p->m02;
        p->m12 = p->m20 * p->m01 - p->m21 * p->m00;
        norm(1);
        
#undef length
#undef norm
        
        p->m03 = -(ex * p->m00 + ey * p->m01 + ez * p->m02);
        p->m13 = -(ex * p->m10 + ey * p->m11 + ez * p->m12);
        p->m23 = -(ex * p->m20 + ey * p->m21 + ez * p->m22);
        
        p->m30 = 0.0f;
        p->m31 = 0.0f;
        p->m32 = 0.0f;
        p->m33 = 1.0f;
}

void
matrix_multiply_4 (matrix4_t *p, matrix4_t *l, matrix4_t *r)
{
        assert(p != l);
        assert(p != r);
        
#define prod(i,j,n) (l->m##i##n * r->m##n##j)
#define dot4(i,j) p->m##i##j = prod(i,j,0)+prod(i,j,1)+prod(i,j,2)+prod(i,j,3)
        
        dot4(0,0); dot4(0,1); dot4(0,2); dot4(0,3);
        dot4(1,0); dot4(1,1); dot4(1,2); dot4(1,3);
        dot4(2,0); dot4(2,1); dot4(2,2); dot4(2,3);
        dot4(3,0); dot4(3,1); dot4(3,2); dot4(3,3);
        
#undef prod
#undef dot4        
}

void
matrix_multiply_vec_4 (float *p, matrix4_t *l, float *r)
{
        assert (p != r);
        p[0] = l->m00 * r[0] + l->m01 * r[1] + l->m02 * r[2] + l->m03 * r[3];
        p[1] = l->m10 * r[0] + l->m11 * r[1] + l->m12 * r[2] + l->m13 * r[3];
        p[2] = l->m20 * r[0] + l->m21 * r[1] + l->m22 * r[2] + l->m23 * r[3];
        p[3] = l->m30 * r[0] + l->m31 * r[1] + l->m32 * r[2] + l->m33 * r[3];
}

void
matrix_ortho_4 (matrix4_t *p, float l, float r, float b, float t, float n,
		float f)
{
        float invdx = (l == r) ? 0.0f : 1.0f / (r - l);
        float invdy = (b == t) ? 0.0f : 1.0f / (t - b);
        float invdz = (n == f) ? 0.0f : 1.0f / (n - f);
        
        p->m00 = - (invdx + invdx);
        p->m01 = 0.0f;
        p->m02 = 0.0f;
        p->m03 = (l + r) * invdx;
        
        p->m10 = 0.0f;
        p->m11 = - (invdy + invdy);
        p->m12 = 0.0f;
        p->m13 = (b + t) * invdy;
        
        p->m20 = 0.0f;
        p->m21 = 0.0f;
        p->m22 = invdz + invdz;
        p->m23 = (f + n) * invdz;
        
        p->m30 = 0.0f;
        p->m31 = 0.0f;
        p->m32 = 0.0f;
        p->m33 = 1.0f;
}

void
matrix_rotate_4 (matrix4_t *p, float angle, float x, float y, float z)
{
        float len = (float) sqrt(x * x + y * y + z * z);
        float inv = (len == 0.0f) ? 0.0f : 1.0f / len;
        
        float theta = angle * DEG_TO_RAD_MULT;
        float s = (float) sin(theta);
        float c = (float) cos(theta);
        
        x *= inv;
        y *= inv;
        z *= inv;
        
        p->m00 = x * x * (1.0f - c) + c;
        p->m01 = y * x * (1.0f - c) - z * s;
        p->m02 = z * x * (1.0f - c) + y * s;
        p->m03 = 0.0f;
        
        p->m10 = x * y * (1.0f - c) + z * s;
        p->m11 = y * y * (1.0f - c) + c;
        p->m12 = z * y * (1.0f - c) - x * s;
        p->m13 = 0.0f;
        
        p->m20 = x * z * (1.0f - c) - y * s;
        p->m21 = y * z * (1.0f - c) + x * s;
        p->m22 = z * z * (1.0f - c) + c;
        p->m23 = 0.0f;
        
        p->m30 = 0.0f;
        p->m31 = 0.0f;
        p->m32 = 0.0f;
        p->m33 = 1.0f;
}

void
matrix_scale_4 (matrix4_t *p, float x, float y, float z)
{
        p->m00 = x; p->m01 = 0.0f; p->m02 = 0.0f; p->m03 = 0.0f;
        p->m10 = 0.0f; p->m11 = y; p->m12 = 0.0f; p->m13 = 0.0f;
        p->m20 = 0.0f; p->m21 = 0.0f; p->m22 = z; p->m23 = 0.0f;
        p->m30 = 0.0f; p->m31 = 0.0f; p->m32 = 0.0f; p->m33 = 1.0f;    
}

void
matrix_translate_4 (matrix4_t *p, float x, float y, float z)
{
        p->m00 = 1.0f; p->m01 = 0.0f; p->m02 = 0.0f; p->m03 = x;
        p->m10 = 0.0f; p->m11 = 1.0f; p->m12 = 0.0f; p->m13 = y;
        p->m20 = 0.0f; p->m21 = 0.0f; p->m22 = 1.0f; p->m23 = z;
        p->m30 = 0.0f; p->m31 = 0.0f; p->m32 = 0.0f; p->m33 = 1.0f;    
}

void
matrix_transpose_4 (matrix4_t *p, matrix4_t *m)
{
        assert(p != m);
        p->m00 = m->m00; p->m01 = m->m10; p->m02 = m->m20; p->m03 = m->m30;
        p->m10 = m->m01; p->m11 = m->m11; p->m12 = m->m21; p->m13 = m->m31;
        p->m20 = m->m02; p->m21 = m->m12; p->m22 = m->m22; p->m23 = m->m32;
        p->m30 = m->m03; p->m31 = m->m13; p->m32 = m->m23; p->m33 = m->m33;
}


void
matrix_perspective_4 (matrix4_t *p, float fov, float aspect, float n, float f)
{
        float h = (float) tan(fov / 2.0f * DEG_TO_RAD_MULT);
        float w = aspect * h;
        matrix_frustum_4(p, -w, w, -h, h, n, f);
}

void
quat_from_angle_axis (quat_t *quat, float angle, float *axis)
{
        float c = (float) cos(angle * (DEG_TO_RAD_MULT / 2.0f));
        float s = (float) sin(angle * (DEG_TO_RAD_MULT / 2.0f));
        float l = (float) sqrt(axis[0] * axis[0] + axis[1] * axis[1] + axis[2] * axis[2]);
        if (l == 0.0f) l = 1.0f;
        quat->x = axis[0] * (s / l);
        quat->y = axis[1] * (s / l);
        quat->z = axis[2] * (s / l);
        quat->w = c;
}

void
quat_to_angle_axis (quat_t *quat, float *angle, float *axis)
{
        float l = (float) sqrt(quat->x * quat->x + quat->y * quat->y + quat->z * quat->z);
        if (l == 0.0f) l = 1.0f;
        *angle = (2.0f / DEG_TO_RAD_MULT) * (float) acos(quat->w); 
        axis[0] = quat->x / l;
        axis[1] = quat->y / l;
        axis[2] = quat->z / l;
        
}
